/* wptSigTreeDlg.cpp - List signatures in a tree view
 *	Copyright (C) 2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptNLS.h"
#include "wptVersion.h"
#include "wptTypes.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptUTF8.h"
#include "wptW32API.h"


/* ID to display signature properties. */
#define _ID_SIGCTX_PROPS 65000


BOOL CALLBACK sigprops_dlg_proc (HWND dlg, UINT msg, 
				 WPARAM wparam, LPARAM lparam);
BOOL CALLBACK keysig_dlg_proc (HWND dlg, UINT msg, 
			       WPARAM wparam, LPARAM lparam);

/* Initialize the signature tree based on the given key @key. */
static void
sigtree_load (HWND dlg, winpt_key_t key)
{
    TVITEM tvi;
    TVINSERTSTRUCT ctx;
    HTREEITEM uid;    
    gpgme_key_sig_t s;
    winpt_key_s signer;
    struct native_uid_s *u;

    memset (&tvi, 0, sizeof (tvi));
    memset (&ctx, 0, sizeof (ctx));
    for (u=key->ext->uids; u; u = u->next) {
	if (u->revoked)
	    continue;
	tvi.pszText = u->uid;
	tvi.state = TVIS_BOLD;
	tvi.mask = TVIF_TEXT;
	tvi.iImage = 0;
	tvi.iSelectedImage = 0;
	ctx.hParent = TVI_ROOT;
	ctx.item = tvi;
	uid = TreeView_InsertItem (dlg, &ctx);
	for (s = u->signatures; s; s=s->next) {
	    memset (&signer, 0, sizeof (signer));
	    if (winpt_get_pubkey (s->keyid+8, &signer))
		continue;
	    tvi.pszText = signer.ext->uids->uid;
	    tvi.mask = TVIF_TEXT|TVIF_IMAGE|TVIF_SELECTEDIMAGE|TVIF_PARAM;
	    tvi.iImage = 1;
	    tvi.iSelectedImage = 1;
	    tvi.lParam = (LPARAM)s;
	    ctx.hParent = uid;
	    ctx.item = tvi;
	    TreeView_InsertItem (dlg, &ctx);
	}
    }
}


/* Associate a signature list based on the icons @ico with
   the treeview control @tree. */
HIMAGELIST
treeview_set_image_list (HWND tree, HICON *ico, DWORD nicons)
{
    HIMAGELIST hil;
    DWORD i;

    hil = ImageList_Create (16, 16, ILC_COLOR8|ILC_MASK, nicons, 1);
    ImageList_SetBkColor (hil, CLR_NONE);
    for (i=0; i < nicons; i++)
	ImageList_AddIcon (hil, ico[i]);    
    TreeView_SetImageList (tree, hil, TVSIL_NORMAL);
    return hil;
}


static void
create_popup (HWND dlg)
{
    HMENU hm;
    POINT p;

    hm = CreatePopupMenu ();
    if (!hm)
	BUG (0);
    insert_menu_item (hm, 0, _ID_SIGCTX_PROPS, _("Signature &Properties"));
    GetCursorPos (&p);
    TrackPopupMenu (hm, 0, p.x, p.y, 0, dlg, NULL);
    DestroyMenu (hm);
}


static void
show_sigprops (HWND dlg)
{
    HWND tree;
    HTREEITEM hti;
    TVITEM tvi;

    memset (&tvi, 0, sizeof (tvi));
    tree = GetDlgItem (dlg, IDC_VKEYSIG_TREE);
    hti = TreeView_GetSelection (tree);
    tvi.mask = TVIF_PARAM;
    tvi.hItem = hti;
    TreeView_GetItem (tree, &tvi);
    if (tvi.lParam != 0)
	DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SIGPROPS, dlg,
	sigprops_dlg_proc, tvi.lParam);
}


/* Dialog box procedure for the tree based signature listing. */
BOOL CALLBACK
sigtree_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static winpt_key_t key;    
    static HIMAGELIST hil;
    HWND tree;
    HICON ico[2];
    NMHDR *nft;
    char inf[300];

    switch (msg) {
    case WM_INITDIALOG:
	key = (winpt_key_t)lparam;
	if (!key)
	    BUG (0);
	ico[0] = LoadIcon (glob_hinst, (LPCTSTR)IDI_USERID);
	ico[1] = LoadIcon (glob_hinst, (LPCTSTR)IDI_KEY_SIG);
	tree = GetDlgItem (dlg, IDC_VKEYSIG_TREE);
	hil = treeview_set_image_list (tree, ico, 2);
	sigtree_load (tree, key);
	_snprintf (inf, sizeof (inf)-1, 
		   _("Signature Tree for \"%s\" (0x%s)"), 
		   key->ext->uids->uid, key->ctx->subkeys->keyid+8);
	SetDlgItemText (dlg, IDC_VKEYSIG_EDIT, _("Edit..."));
	SetWindowText (dlg, inf);
	SetForegroundWindow (dlg);
	break;

    case WM_DESTROY:
	ImageList_Destroy (hil);
	break;

    case WM_NOTIFY:
	nft = (NMHDR*)lparam;
	if (nft->code == NM_DBLCLK)
	    show_sigprops (dlg);
	else if (nft->code == NM_RCLICK)
	    create_popup (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case _ID_SIGCTX_PROPS:
	    show_sigprops (dlg);
	    break;

	case IDOK:
	    EndDialog (dlg, 1);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, 0);
	    break;

	case IDC_VKEYSIG_EDIT:
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_KEYSIG, dlg,
			    keysig_dlg_proc, (LPARAM)key);
	}
	break;
    }

    return FALSE;
}
